"""Abstractions for dataset file formats.

Handles reading in the data and transforming data types to generic metadata types.
"""

from __future__ import annotations

import pathlib  # noqa: TCH003 import is needed for docs build
import re
import typing as t
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

import pandas as pd
from datadoc_model.model import DataType
from datadoc_model.model import LanguageStringType
from datadoc_model.model import LanguageStringTypeItem
from datadoc_model.model import Variable
from pyarrow import parquet as pq

from dataset.utility.enums import SupportedLanguages

if TYPE_CHECKING:
    import pyarrow as pa
    from cloudpathlib import CloudPath

KNOWN_INTEGER_TYPES = (
    "int",
    "int_",
    "int8",
    "int16",
    "int32",
    "int64",
    "integer",
    "long",
    "uint",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
)

KNOWN_FLOAT_TYPES = (
    "double",
    "float",
    "float_",
    "float16",
    "float32",
    "float64",
    "decimal",
    "number",
    "numeric",
    "num",
)

KNOWN_STRING_TYPES = (
    "string",
    "str",
    "char",
    "varchar",
    "varchar2",
    "text",
    "txt",
    "bytes",
)

KNOWN_DATETIME_TYPES = (
    "timestamp",
    "timestamp[us]",
    "timestamp[ns]",
    "datetime64",
    " datetime64[ns]",
    " datetime64[us]",
    "date",
    "datetime",
    "time",
)

KNOWN_BOOLEAN_TYPES = ("bool", "bool_", "boolean")


TYPE_CORRESPONDENCE: list[tuple[tuple[str, ...], DataType]] = [
    (KNOWN_INTEGER_TYPES, DataType.INTEGER),
    (KNOWN_FLOAT_TYPES, DataType.FLOAT),
    (KNOWN_STRING_TYPES, DataType.STRING),
    (KNOWN_DATETIME_TYPES, DataType.DATETIME),
    (KNOWN_BOOLEAN_TYPES, DataType.BOOLEAN),
]
TYPE_MAP: dict[str, DataType] = {}
for concrete_type, abstract_type in TYPE_CORRESPONDENCE:
    TYPE_MAP.update({c: abstract_type for c in concrete_type})

TDatasetParser = t.TypeVar("TDatasetParser", bound="DatasetParser")


class DatasetParser(ABC):
    """Abstract Base Class for all Dataset parsers.

    Implements:
    - A static factory method to get the correct implementation for each file extension.
    - A static method for data type conversion.

    Requires implementation by subclasses:
    - A method to extract variables (columns) from the dataset, so they may be documented.
    """

    def __init__(self, dataset: pathlib.Path | CloudPath) -> None:
        """Initialize for a given dataset."""
        self.dataset = dataset

    @staticmethod
    def for_file(dataset: pathlib.Path | CloudPath) -> DatasetParser:
        """Return the correct subclass based on the given dataset file."""
        supported_file_types: dict[
            str,
            type[DatasetParser],
        ] = {
            ".parquet": DatasetParserParquet,
            ".sas7bdat": DatasetParserSas7Bdat,
            ".parquet.gzip": DatasetParserParquet,
        }
        file_type = "Unknown"
        try:
            file_type = dataset.suffix
            # Gzipped parquet files can be read with DatasetParserParquet
            match = re.search(r"(.parquet.gzip)", str(dataset).lower())
            file_type = ".parquet.gzip" if match else file_type
            # Extract the appropriate reader class from the SUPPORTED_FILE_TYPES dict and return an instance of it
            reader = supported_file_types[file_type](dataset)
        except IndexError as e:
            # Thrown when just one element is returned from split, meaning there is no file extension supplied
            msg = f"Could not recognise file type for provided {dataset = }. Supported file types are: {', '.join(supported_file_types.keys())}"
            raise FileNotFoundError(
                msg,
            ) from e
        except KeyError as e:
            # In this case the file type is not supported, so we throw a helpful exception
            msg = f"{file_type = } is not supported. Please open one of the following supported files types: {', '.join(supported_file_types.keys())} or contact the maintainers to request support."
            raise NotImplementedError(
                msg,
            ) from e
        else:
            return reader

    @staticmethod
    def transform_data_type(data_type: str) -> DataType | None:
        """Transform a concrete data type to an abstract data type.

        In statistical metadata, one is not interested in how the data is
        technically stored, but in the meaning of the data type. Because of
        this, we transform known data types to their abstract metadata
        representations.

        If we encounter a data type we don't know, we just ignore it and let
        the user handle it in the GUI.

        Arguments:
            data_type: The concrete data type to map.
        """
        return TYPE_MAP.get(data_type.lower(), None)

    @abstractmethod
    def get_fields(self) -> list[Variable]:
        """Abstract method, must be implemented by subclasses."""


class DatasetParserParquet(DatasetParser):
    """Concrete implementation for parsing parquet files."""

    def __init__(self, dataset: pathlib.Path | CloudPath) -> None:
        """Call the super init method for initialization.

        Args:
            dataset: Path to the dataset to parse.
        """
        super().__init__(dataset)

    def get_fields(self) -> list[Variable]:
        """Extract the fields from this dataset."""
        with self.dataset.open(mode="rb") as f:
            schema: pa.Schema = pq.read_schema(f)  # type: ignore [arg-type]
        return [
            Variable(
                short_name=data_field.name.strip(),
                data_type=self.transform_data_type(str(data_field.type)),
            )
            for data_field in schema
            if data_field.name
            != "__index_level_0__"  # Index columns should not be documented
        ]


class DatasetParserSas7Bdat(DatasetParser):
    """Concrete implementation for parsing SAS7BDAT files."""

    def __init__(self, dataset: pathlib.Path | CloudPath) -> None:
        """Call the super init method for initialization.

        Args:
            dataset: Path to the dataset to parse.
        """
        super().__init__(dataset)

    def get_fields(self) -> list[Variable]:
        """Extract the fields from this dataset."""
        fields = []
        with self.dataset.open(mode="rb") as f:
            # Use an iterator to avoid reading in the entire dataset
            sas_reader = pd.read_sas(f, format="sas7bdat", iterator=True)

            # Get the first row from the iterator
            try:
                row = next(sas_reader)
            except StopIteration as e:
                msg = f"Could not read data from {self.dataset}"
                raise RuntimeError(msg) from e

        # Get all the values from the row and loop through them
        for i, v in enumerate(row.to_numpy().tolist()[0]):
            fields.append(
                Variable(
                    short_name=sas_reader.columns[i].name,  # type: ignore [attr-defined]
                    # Assume labels are defined in the default language (NORSK_BOKMÅL)
                    # If this is not correct, the user may fix it via the UI
                    name=LanguageStringType(
                        [
                            LanguageStringTypeItem(
                                languageCode=SupportedLanguages.NORSK_BOKMÅL.value,
                                languageText=sas_reader.columns[  # type: ignore [attr-defined]
                                    i
                                ].label,
                            ),
                        ],
                    ),
                    # Access the python type for the value and transform it to a DataDoc Data type
                    data_type=self.transform_data_type(type(v).__name__.lower()),  # type: ignore  # noqa: PGH003
                ),
            )

        return fields



"""Enumerations used in Datadoc."""

from __future__ import annotations

from enum import Enum


class DaplaRegion(str, Enum):
    """Dapla platforms/regions."""

    DAPLA_LAB = "DAPLA_LAB"
    BIP = "BIP"
    ON_PREM = "ON_PREM"
    CLOUD_RUN = "CLOUD_RUN"


class DaplaService(str, Enum):
    """Dapla services."""

    DATADOC = "DATADOC"
    JUPYTERLAB = "JUPYTERLAB"
    VS_CODE = "VS_CODE"
    R_STUDIO = "R_STUDIO"
    KILDOMATEN = "KILDOMATEN"


class SupportedLanguages(str, Enum):
    """The list of languages metadata may be recorded in.

    Reference: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry
    """

    NORSK_BOKMÅL = "nb"
    NORSK_NYNORSK = "nn"
    ENGLISH = "en"



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

import bs4
import requests
from bs4 import BeautifulSoup
from bs4 import ResultSet

from dataset.external_sources.external_sources import GetExternalSource
from dataset.utility.enums import SupportedLanguages

if TYPE_CHECKING:
    from concurrent.futures import ThreadPoolExecutor

logger = logging.getLogger(__name__)


@dataclass
class Subject:
    """Base class for Primary and Secondary subjects.

    A statistical subject is a related grouping of statistics.
    """

    titles: dict[str, str]
    subject_code: str

    def get_title(self, language: SupportedLanguages) -> str:
        """Get the title in the given language."""
        try:
            return self.titles[
                (
                    # Adjust to language codes in the StatisticSubjectMapping structure.
                    "no"
                    if language
                    in [
                        SupportedLanguages.NORSK_BOKMÅL,
                        SupportedLanguages.NORSK_NYNORSK,
                    ]
                    else "en"
                )
            ]
        except KeyError:
            logger.exception(
                "Could not find title for subject %s  and language: %s",
                self,
                language.name,
            )
            return ""


@dataclass
class SecondarySubject(Subject):
    """Data structure for secondary subjects or 'delemne'."""

    statistic_short_names: list[str]


@dataclass
class PrimarySubject(Subject):
    """Data structure for primary subjects or 'hovedemne'."""

    secondary_subjects: list[SecondarySubject]


class StatisticSubjectMapping(GetExternalSource):
    """Provide mapping between statistic short name and primary and secondary subject."""

    def __init__(
        self,
        executor: ThreadPoolExecutor,
        source_url: str | None,
    ) -> None:
        """Retrieve the statistical structure document from the given URL.

        Initializes the mapping based on values in the statistical structure document sourced at `source_url`.

        Args:
            executor: The ThreadPoolExecutor which will run the job of fetching the statistical structure document.
            source_url: The URL from which to fetch the statistical structure document.
        """
        self.source_url = source_url

        self._statistic_subject_structure_xml: ResultSet | None = None

        self._primary_subjects: list[PrimarySubject] = []

        super().__init__(executor)

    def get_secondary_subject(self, statistic_short_name: str | None) -> str | None:
        """Looks up the secondary subject for the given statistic short name in the mapping dict.

        Returns the secondary subject string if found, else None.
        """
        for p in self.primary_subjects:
            for s in p.secondary_subjects:
                if statistic_short_name in s.statistic_short_names:
                    logger.debug("Got %s from %s", s, statistic_short_name)
                    return s.subject_code

        logger.debug("No secondary subject found for %s", statistic_short_name)
        return None

    @staticmethod
    def _extract_titles(titles_xml: bs4.element.Tag) -> dict[str, str]:
        titles = {}
        for title in titles_xml.find_all("tittel"):
            titles[title["sprak"]] = title.text
        return titles

    def _fetch_data_from_external_source(self) -> ResultSet | None:
        """Fetch statistical structure document from source_url.

        Returns a BeautifulSoup ResultSet.
        """
        try:
            url = str(self.source_url)
            response = requests.get(url, timeout=30)
            response.encoding = "utf-8"
            logger.debug("Got response %s from %s", response, url)
            soup = BeautifulSoup(response.text, features="xml")
            return soup.find_all("hovedemne")
        except requests.exceptions.RequestException:
            logger.exception(
                "Exception while fetching statistical structure ",
            )
            return None

    def _parse_statistic_subject_structure_xml(
        self,
        statistical_structure_xml: ResultSet,
    ) -> list[PrimarySubject]:
        primary_subjects: list[PrimarySubject] = []
        for p in statistical_structure_xml:
            secondary_subjects: list[SecondarySubject] = [
                SecondarySubject(
                    self._extract_titles(s.titler),
                    s["emnekode"],
                    [statistikk["kortnavn"] for statistikk in s.find_all("Statistikk")],
                )
                for s in p.find_all("delemne")
            ]

            primary_subjects.append(
                PrimarySubject(
                    self._extract_titles(p.titler),
                    p["emnekode"],
                    secondary_subjects,
                ),
            )
        return primary_subjects

    @property
    def primary_subjects(self) -> list[PrimarySubject]:
        """Getter for primary subjects."""
        if not self._primary_subjects:
            self._parse_xml_if_loaded()
            logger.debug("Got %s primary subjects", len(self._primary_subjects))
        return self._primary_subjects

    def _parse_xml_if_loaded(self) -> bool:
        """Checks if the xml is loaded, then parses the xml if it is loaded.

        Returns `True` if it is loaded and parsed.
        """
        if self.check_if_external_data_is_loaded():
            self._statistic_subject_structure_xml = self.retrieve_external_data()

            if self._statistic_subject_structure_xml is not None:
                self._primary_subjects = self._parse_statistic_subject_structure_xml(
                    self._statistic_subject_structure_xml,
                )
                logger.debug(
                    "Thread finished. Parsed %s primary subjects",
                    len(self._primary_subjects),
                )
                return True
            logger.warning("Thread is not done. Cannot parse xml.")
        return False



"""Document dataset."""

from datadoc_model import model

from .core import Datadoc
from .dapla_dataset_path_info import DaplaDatasetPathInfo
from .model_validation import ObligatoryDatasetWarning
from .model_validation import ObligatoryVariableWarning
from .utility.enums import DaplaRegion
from .utility.enums import DaplaService
from .utility.enums import SupportedLanguages



"""Handle validation for metadata with pydantic validators and custom warnings."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING
from typing import TextIO

from datadoc_model import model
from pydantic import model_validator
from typing_extensions import Self

from dataset.utility.constants import DATE_VALIDATION_MESSAGE
from dataset.utility.constants import NUM_OBLIGATORY_DATASET_FIELDS
from dataset.utility.constants import NUM_OBLIGATORY_VARIABLES_FIELDS
from dataset.utility.constants import OBLIGATORY_METADATA_WARNING
from dataset.utility.utils import get_missing_obligatory_dataset_fields
from dataset.utility.utils import get_missing_obligatory_variables_fields
from dataset.utility.utils import get_timestamp_now
from dataset.utility.utils import incorrect_date_order
from dataset.utility.utils import num_obligatory_dataset_fields_completed
from dataset.utility.utils import num_obligatory_variables_fields_completed
from dataset.utility.utils import set_variables_inherit_from_dataset

if TYPE_CHECKING:
    from datetime import datetime

logger = logging.getLogger(__name__)


class ValidateDatadocMetadata(model.DatadocMetadata):
    """Class that inherits from DatadocMetadata, providing additional validation."""

    @model_validator(mode="after")
    def check_date_order(self) -> Self:
        """Validate the order of date fields.

        Check that dataset and variable date fields `contains_data_from` and
        `contains_data_until` are in chronological order.

        Mode: This validator runs after other validation.

        Returns:
            The instance of the model after validation.

        Raises:
            ValueError: If `contains_data_until` date is earlier than
                `contains_data_from date`.
        """
        if self.dataset is not None and incorrect_date_order(
            self.dataset.contains_data_from,
            self.dataset.contains_data_until,
        ):
            raise ValueError(DATE_VALIDATION_MESSAGE)
        if self.variables is not None:
            for v in self.variables:
                if incorrect_date_order(v.contains_data_from, v.contains_data_until):
                    raise ValueError(DATE_VALIDATION_MESSAGE)
        return self

    @model_validator(mode="after")
    def check_metadata_created_date(self) -> Self:
        """Ensure `metadata_created_date` is set for the dataset.

        Sets the current timestamp if `metadata_created_date` is None.

        Mode: This validator runs after other validation.

        Returns:
            The instance of the model after validation.
        """
        timestamp: datetime = get_timestamp_now()  # --check-untyped-defs
        if self.dataset is not None and self.dataset.metadata_created_date is None:
            self.dataset.metadata_created_date = timestamp
        return self

    @model_validator(mode="after")
    def check_inherit_values(self) -> Self:
        """Inherit values from dataset to variables if not set.

        Sets values for 'data source', 'temporality type', 'contains data from',
        and 'contains data until' if they are None.

        Mode: This validator runs after other validation.

        Returns:
            The instance of the model after validation.
        """
        if self.variables and self.dataset is not None:
            set_variables_inherit_from_dataset(self.dataset, self.variables)
        return self

    @model_validator(mode="after")
    def check_obligatory_dataset_metadata(self) -> Self:
        """Check obligatory dataset fields and issue a warning if any are missing.

        Mode:
            This validator runs after other validation.

        Returns:
            The instance of the model after validation.

        Raises:
            ObligatoryDatasetWarning: If not all obligatory dataset metadata fields
                are filled in.
        """
        if (
            self.dataset is not None
            and num_obligatory_dataset_fields_completed(
                self.dataset,
            )
            != NUM_OBLIGATORY_DATASET_FIELDS
        ):
            warnings.warn(
                f"{OBLIGATORY_METADATA_WARNING} {get_missing_obligatory_dataset_fields(self.dataset)}",
                ObligatoryDatasetWarning,
                stacklevel=2,
            )
            logger.warning(
                "Type warning: %s.%s %s",
                ObligatoryDatasetWarning,
                OBLIGATORY_METADATA_WARNING,
                get_missing_obligatory_dataset_fields(self.dataset),
            )

        return self

    @model_validator(mode="after")
    def check_obligatory_variables_metadata(self) -> Self:
        """Check obligatory variable fields and issue a warning if any are missing.

        Mode:
            This validator runs after other validation.

        Returns:
            The instance of the model after validation.

        Raises:
            ObligatoryVariableWarning: If not all obligatory variable metadata fields
                are filled in.
        """
        if self.variables is not None and num_obligatory_variables_fields_completed(
            self.variables,
        ) != (NUM_OBLIGATORY_VARIABLES_FIELDS * len(self.variables)):
            warnings.warn(
                f"{OBLIGATORY_METADATA_WARNING} {get_missing_obligatory_variables_fields(self.variables)}",
                ObligatoryVariableWarning,
                stacklevel=2,
            )
            logger.warning(
                "Type warning: %s.%s %s",
                ObligatoryVariableWarning,
                OBLIGATORY_METADATA_WARNING,
                get_missing_obligatory_variables_fields(self.variables),
            )

        return self


class ValidationWarning(UserWarning):
    """Custom warning for validation purposes."""


class ObligatoryDatasetWarning(UserWarning):
    """Custom warning for checking obligatory metadata for dataset."""


class ObligatoryVariableWarning(UserWarning):
    """Custom warning for checking obligatory metadata for variables."""


def custom_warning_handler(  # noqa: PLR0913 remove fields causes incompatible types
    message: Warning | str,
    category: type[Warning],
    filename: str,
    lineno: int,
    file: TextIO | None = None,  # noqa: ARG001 remove causes incompatible types
    line: str | None = None,  # noqa: ARG001 remove causes incompatible types
) -> None:
    """Handle warnings."""
    print(  # noqa: T201
        f"Warning: {message}, Category: {category.__name__}, Filename: {filename}, Line: {lineno}",
    )


warnings.showwarning = custom_warning_handler
warnings.simplefilter("always")
